/***********************************************************************
 * Texture Denoising
 * 
 * compile: c++ -std=c++11 `Magick++-config --cxxflags --cppflags` -I./ -O2 -o denoise texturedenoising.cpp `Magick++-config --ldflags --libs` -lexternal-library-qpbo-shared -lexternal-library-maxflow-shared
 * 
 * run: ./denoise [lambda] [sigma] [image filename]
 ************************************************************************/

#include <Magick++.h> 
#include <iostream>
#include <cstdio>
#include <random>
#include <cmath>
#include <cassert>

#include <opengm/opengm.hxx>
#include <opengm/graphicalmodel/graphicalmodel.hxx>
#include <opengm/operations/adder.hxx>
#include <opengm/operations/minimizer.hxx>
#include <opengm/inference/graphcut.hxx>
#include <opengm/inference/auxiliary/minstcutboost.hxx>
#include <opengm/inference/external/qpbo.hxx>

using namespace Magick;

//*****************************************
//** Typedefs
//*****************************************
// type used for values
typedef double ValueType;
// type used for indexing nodes and factors (default : size_t)
typedef size_t IndexType;
// type used for labels (default : size_t)
typedef size_t LabelType;
// operation used to combine terms
typedef opengm::Adder OpType;
// shortcut for explicite function
typedef opengm::ExplicitFunction<ValueType,IndexType,LabelType> ExplicitFunction;
// list of all function the model cal use (this trick avoids virtual methods) - here only one
typedef opengm::meta::TypeListGenerator<ExplicitFunction>::type FunctionTypeList;
// type used to define the feasible statespace
typedef opengm::DiscreteSpace<IndexType, LabelType> SpaceType;
// type of the model
typedef opengm::GraphicalModel<ValueType,OpType,FunctionTypeList,SpaceType> Model;
// type of the function identifier
typedef Model::FunctionIdentifier FunctionIdentifier;

typedef opengm::MinSTCutBoost<size_t, double, opengm::PUSH_RELABEL> MinSTCutType;
typedef opengm::GraphCut<Model, opengm::Minimizer, MinSTCutType> MinCut;
typedef opengm::external::QPBO<Model> QPBO;

//*****************************************
//** Declare functions 
//*****************************************
void addNoise(const Image& _image, Image &_noisyImage, double _sigma);
double imageDiff(const Image& _image1, const Image& _image2);

//*****************************************
//** main
//*****************************************
int main(int argc, char** argv) {
    InitializeMagick(*argv);
    
    //*****************************************
    //** Parameters
    //*****************************************
    double lambda = 1.0;
    double sigma = 0.7;
    std::string imageFilename = "image3.png";
    
    if (argc>1)
    {
        lambda = std::stod(std::string(argv[1]));
    }
    if (argc>2)
    {
        sigma = std::stod(std::string(argv[2]));
    }
    if (argc>3)
    {
        imageFilename = std::string(argv[3]);
    }
    
    try{
        //*****************************************
        //** Load Image and create Noisy image
        //*****************************************
        
        Image image;
        image.read(imageFilename);

        int W = image.size().width();
        int H = image.size().height();

        Image noisyImage;
        addNoise(image, noisyImage,sigma);
        
        // Show noisy image
        Image tempImage = noisyImage;
        tempImage.display();
        //*****************************************
        //** Construct Model
        //*****************************************
        std::cout << "Start building the model ... "<<std::endl;
        // Build empty Model
        LabelType numLabel = 2;
        std::vector<LabelType> numbersOfLabels(W*H,numLabel);
        //construct two models: "gm" includes all pairwise terms and "gmSubmodular" only includes the submodular pairwise terms.
        Model gm(SpaceType(numbersOfLabels.begin(), numbersOfLabels.end()));
        Model gmSubmodular(SpaceType(numbersOfLabels.begin(), numbersOfLabels.end()));
	
	TODO add you code here to construct the graphical models "gm" and "gmSubmodular"
	
        /** inference computation (optimization) starts here
            You can use either the ICM algorithm or TRWSi. Both will optimize the submodular energy of the exercise.
            */
        QPBO qpbo(gm);
        qpbo.infer();
        
        std::vector<bool> optimalVariables;
        bool partialOptimality = qpbo.partialOptimality(optimalVariables);
        
        std::cout << "value: " << qpbo.value() << " bound: " << qpbo.bound() << std::endl;
        std::cout << "partial optimality: " << partialOptimality << std::endl;
        std::vector<LabelType> l_QPBO;
        qpbo.arg(l_QPBO);
        Image qpboImage = image;
        for (int y=0; y<H; ++y)
        {
            for (int x=0; x<W; ++x)
            {
                int linearIndex = x+W*y;
                ColorGray c = qpboImage.pixelColor(x,y);
                c.shade((double)l_QPBO[linearIndex]);
                if (optimalVariables[linearIndex])
                    qpboImage.pixelColor(x,y,c);
                else
                    qpboImage.pixelColor(x,y,ColorRGB(1,0,0));
            }
        }
        
        double errorQPBO = imageDiff(qpboImage, image);
        std::cout << "Error :  " << errorQPBO << "   Ratio : " << (errorQPBO/(W*H));
        std::cout  << std::endl;
        // Show Graph Cut image
        tempImage = qpboImage;
        tempImage.display();

        // Infer with Graph Cut
        std::cout << "Start GraphCut inference ... " <<std::endl;
        MinCut::Parameter para(1000000); //scale capacities, only needed if we use integer capacities
        MinCut minCut(gmSubmodular,para);
        minCut.infer();
        std::vector<LabelType> l_MinCut;
        minCut.arg(l_MinCut);
        std::cout << "Energy :  "<<minCut.value() << std::endl; 
        Image graphCutImage = image;
        for (int y=0; y<H; ++y)
        {
            for (int x=0; x<W; ++x)
            {
                int linearIndex = x+W*y;
                ColorGray c = graphCutImage.pixelColor(x,y);
                c.shade((double)l_MinCut[linearIndex]);
                graphCutImage.pixelColor(x,y,c);
            }
        }
        
        double errorGC = imageDiff(graphCutImage, image);
        std::cout << "Error :  " << errorGC << "   Ratio : " << (errorGC/(W*H));
        std::cout  << std::endl;
        // Show Graph Cut image
        tempImage = graphCutImage;
        tempImage.display();

    }
    catch( Exception &error_ ) 
    { 
        std::cout << "Caught exception: " << error_.what() << std::endl; 
        return 1; 
    } 


}


/***********************************************************************
 * Implementation other functions
 * 
 ************************************************************************/

void addNoise(const Image& _image, Image &_noisyImage, double _sigma)
{
    std::default_random_engine generator(2);
    std::normal_distribution<double> distribution(0.0,_sigma);
    Geometry imageSize = _image.size();
    _noisyImage = _image;
    //std::cout << "(w,h)=(" << imageSize.width() << "," << imageSize.height() << ")"<< std::endl;
    for (int y=0; y<imageSize.height(); ++y)
    {
        for (int x=0; x<imageSize.width(); ++x)
        {
            double number = distribution(generator);
            ColorGray c = _noisyImage.pixelColor(x,y);
            c.shade(std::max(0.0,std::min(1.0,number+c.shade())));
            _noisyImage.pixelColor(x,y,c);
        }
    }
}

double imageDiff(const Image& _image1, const Image& _image2)
{
    double result = 0;
    Geometry imageSize = _image1.size();
    for (int y=0; y<imageSize.height(); ++y)
    {
        for (int x=0; x<imageSize.width(); ++x)
        {
            ColorGray c1 = _image1.pixelColor(x,y);
            ColorGray c2 = _image2.pixelColor(x,y);
            result += std::fabs(c1.shade()-c2.shade());
        }
    }
    return result;
}
    
